from pypaq.lipytools.pylogger import get_pylogger
from pypaq.mpython.mptools import sys_res_nfo
from pypaq.exception import PyPaqException
from typing import Union, List

DevicesPypaq: Union[None, float, str, List[Union[None,float,str]]] = 1.0


def get_devices(
        devices: DevicesPypaq=  1.0,
        logger=                 None,
        loglevel=               20,
) -> List[None]:
    """ resolves representation given with devices (DevicesTorchness) """

    if not logger:
        logger = get_pylogger(name='get_devices', level=loglevel)

    if type(devices) is not list:
        devices = [devices]

    if not devices:
        raise PyPaqException('no devices given')

    cpu_count = sys_res_nfo()['cpu_count']
    logger.debug(f'got {cpu_count} CPU devices in a system')

    devices_base = []
    for d in devices:

        known_device = False

        if d is None:
            known_device = True
            devices_base.append(d)

        if type(d) is float:
            known_device = True
            if d < 0.0: d = 0.0
            if d > 1.0: d = 1.0
            cpu_count_f = round(cpu_count * d)
            if cpu_count_f < 1: cpu_count_f = 1
            devices_base += [None] * cpu_count_f

        if type(d) is str:
            if d == 'all':
                known_device = True
                devices_base += [None] * cpu_count
            if 'cpu' in d:
                known_device = True
                devices_base.append(None)

        if not known_device:
            msg = f'unknown (not valid?) device given: {d}'
            logger.error(msg)
            raise PyPaqException(msg)

    return devices_base


import random
import string
import time
from typing import List, Any, Optional

from pypaq.exception import PyPaqException
from pypaq.pytypes import NUM
from pypaq.lipytools.moving_average import MovAvg

LINE_UP = '\033[1A'
LINE_CLEAR = '\x1b[2K'


def nice_scin(num:NUM, precision:int=1, replace_zero:bool=True, add_plus:bool=False) -> str:
    """short (compressed) scientific notation for numbers

    examples:
        0.1                  ->: 1.0e-1
        0.13124              ->: 1.3e-1
        0.45738683456        ->: 4.6e-1
        0.0021342            ->: 2.1e-3
        1.234e-06            ->: 1.2e-6
        0.0                  ->: 0.0e0
        1.3453               ->: 1.3e0
        1.9999999            ->: 2.0e0
        123.12345            ->: 1.2e2
        1                    ->: 1.0e0
        2                    ->: 2.0e0
        100.0                ->: 1.0e2
        -0.0                 ->: 0.0e0
        -1                   ->: -1.0e0
        -2                   ->: -2.0e0
        -999                 ->: -1.0e3
        9999                 ->: 1.0e4
        -99999999            ->: -1.0e8
        99999999             ->: 1.0e8
    """

    # -0.0 case
    if num == 0:
        num = 0

    sh = f'{num:.{precision}e}'

    if replace_zero:
        sh = sh.replace('+0','+')
        if not add_plus:
            sh = sh.replace('+','')
        sh = sh.replace('-0','-')

    if add_plus and num >= 0:
        sh = '+' + sh

    return sh


def nice_float_pad(
        num: float,
        width: int=     7,
        add_plus: bool= False,
        fill: bool=     True,
) -> str:
    """returns nice string from float, always of a given width, padded with spaces
    width should be >= 5

    examples:
        0.1                  -> 0.1     (7)
        0.13124              -> 0.13124 (7)
        0.45738683456        -> 0.45738 (7)
        0.0021342            -> 0.00213 (7)
        1.234e-06            -> 1.23e-6 (7)
        0.0                  -> 0       (7)
        1.3453               -> 1.3453  (7)
        1.9999999            -> 1.99999 (7)
        123.12345            -> 123.123 (7)
        1                    -> 1       (7)
        2                    -> 2       (7)
        100.0                -> 100     (7)
        -0.0                 -> 0       (7)
        -1                   -> -1      (7)
        -2                   -> -2      (7)
        -999                 -> -999    (7)
        9999                 -> 9999    (7)
        -888888888           -> -8.89e8 (7)
        888888888            -> 8.889e8 (7)
    """

    if width < 5:
        raise ValueError('width must be >= 5')

    # 1e2 case
    if round(num) == num:
        num = int(num)

    if 1e4 > num > 1e-4 or -1e4 < num < -1e-4 or num == 0:
        fstr = str(num)[:width]
    else:
        precision = width - 4
        fstr = '-' * (width+1)
        while len(fstr) > width:
            fstr = nice_scin(num, precision=precision, replace_zero=True, add_plus=add_plus)
            precision -= 1

    if fill and len(fstr)<width:
        fstr += ' '*(width-len(fstr))

    return fstr


def nice_float_width(
        num: float,
        width: int=     4,
        add_plus: bool= False,
) -> str:
    """returns nice string from float from range (-(10**(width-1));10**width)

    example (width = 4):
        0.1                  -> .100
        0.13124              -> .131
        0.4573868            -> .457
        0.0021342            -> .002
        1.234e-06            -> .000
        0.0                  -> .000
        1.3453               -> 1.34
        1.9999999            -> 2.00
        123.123              -> 123.
        1                    -> 1.00
        2                    -> 2.00
        100.0                -> 100.
        -0.0                 -> -.00
        -1                   -> -1.0
        -2                   -> -2.0
        -999                 -> -999
        9999                 -> 9999
    """
    if num >= 10**width or -(10**(width-1)) >= num:
        raise ValueError(f'this value num: {num} is out of supported range!')
    num_str = f"{num:.{width+1}f}"
    if add_plus and num >= 0:
        num_str = '+' + num_str
    if num_str[0] == '0':
        num_str = num_str[1:]
    if num_str.startswith('-0'):
        num_str = '-' + num_str[2:]
    num_str = num_str[:width]
    while len(num_str) < width:
        num_str += '0'
    return num_str


def stamp(
        year: bool=             False,
        month: bool=            True,
        day: bool=              True,
        hour: bool=             True,
        minutes: bool=          True,
        letters: Optional[int]= 3,
        separator: str=         '_',
) -> str:
    """ prepares timestamp string """

    random.seed(time.time())

    time_format = ''
    if year:            time_format += '%y'
    if month:           time_format += '%m'
    if day:             time_format += '%d'
    if time_format and (hour or minutes): time_format += separator
    if hour:            time_format += '%H'
    if minutes:         time_format += '%M'

    stp = ''
    if time_format:
        stp = time.strftime(time_format)

    if letters:
        if stp: stp += separator
        stp += ''.join([random.choice(string.ascii_letters) for _ in range(letters)])

    return stp


def list_str(ls: List[Any], limit:Optional[int]=200) -> str:
    """ nice string of given list """
    lstr = [str(e) for e in ls]
    lstr = '; '.join(lstr)
    if limit: lstr = lstr[:limit]
    return lstr


def print_nested_dict(
        d: dict,
        ind_scale: int=     2,
        line_limit: int=    150,
) -> None:
    """ prints nice string of nested dict """

    types = {
        int:        'int',
        float:      'float',
        bool:       'bool',
        type(None): 'NoneType',
        dict:       'dict',
        list:       'list',
        tuple:      'tuple',
        str:        'str'}

    types_len = ['dict','list','tuple','str']

    def __prn_root(root:dict, ind:int):

        spacer = ' ' * ind * ind_scale

        for key in sorted(list(root.keys())):

            value = root[key]
            tp = types.get(type(value),None)
            value_len_str = str(len(value)) if tp in types_len else ''

            if tp == 'list':
                sub_tp = list(set([type(e) for e in value]))
                if len(sub_tp) == 1:
                    sub_tp = types.get(sub_tp[0],None)
                    if sub_tp:
                        tp = f'{tp}[{sub_tp}]'

            val_line = ''
            if tp != 'dict':
                val_line = repr(value)
                if line_limit and len(val_line)>line_limit:
                    val_line = f'{val_line[:line_limit]} ..'
                val_line = f' : {val_line}'

            value_len_nfo = f'.{value_len_str}' if value_len_str else ''
            type_nfo = f' [{tp}{value_len_nfo}]' if tp else ''
            print(f'{spacer}{key}{type_nfo}{val_line}')

            if tp == 'dict':
                __prn_root(root=root[key], ind=ind+1)

    __prn_root(root=d, ind=0)


def printover(sth, clear:int=10) -> None:
    """ prints line over line """
    cls = '' + ' ' * clear
    print(f'\r{sth}{cls}', end='', flush=True)


def printover_terminal(sth) -> None:
    """ prints line over line (terminal alternative) """
    print(sth)
    print(LINE_UP, end=LINE_CLEAR)


class ProgBar:
    """ ProgBar - terminal progress bar """

    def __init__(
            self,
            total: NUM,
            name: Optional[str]=            None,
            length: int=                    20,
            fill: str=                      '█',
            show_fract: bool=               True,
            show_speed_avg: bool=           True,
            show_speed_cur: bool=           True,
            show_eta: bool=                 True,
            guess_speed: float=             10.0,
            refresh_delay: Optional[float]= 1,
            logger=                         None,
    ):
        """ speed_avg is an average speed smoothened with moving average,
        speed_cur is a current speed smoothened with moving average,

        total: task size
        length: bar char length
        fill: char to fill the bar
        show_fract: fraction of task
        show_speed_avg: speed moving average
        show_speed_cur: speed current (also moving average but with lower factor)
        show_eta: ETA
        guess_speed: expected speed value (initial guess), helps to set proper params for the update
        refresh_delay: (sec) delay between refreshes
        """
        self.total = total
        self.name = name or ''
        self.length = length
        self.fill = fill
        self.show_fract = show_fract
        self.show_speed_avg = show_speed_avg
        self.show_speed_cur = show_speed_cur and self.show_speed_avg
        self.show_eta = show_eta
        self.refresh_delay = refresh_delay
        self.logger = logger

        f = min(0.5,max(0.01, 1/guess_speed))
        self.speed_a = MovAvg(factor=f) # tot mavg
        self.speed_c = MovAvg(factor=f) # current mavg
        self.n_prev = 0
        self.inc_cached = 0
        self.start_time = time.time()
        self.time_prev = self.start_time

    @staticmethod
    def _speed_to_str(s) -> str:
        if s > 1:
            if s > 100: return f'{int(s)}/s'
            else:       return f'{s:.1f}/s'
        else:           return f'{s:.3f}/s'

    @staticmethod
    def _time_to_str(t) -> str:
        if t > 4000:    return f'{t / 60 / 60:.1f}h'
        else:
            if t > 100: return f'{t / 60:.1f}m'
            else:       return f'{t:.1f}s'

    def __call__(self, n:NUM, prefix:str='', suffix:str=''):

        if n < self.n_prev:
            raise PyPaqException('ProgBar cannot step back with progress (n < prev)')

        if n > self.n_prev:

            time_current = time.time()
            time_passed_prev = time_current - self.time_prev
            if self.refresh_delay is None or time_passed_prev >= self.refresh_delay or n >= self.total:

                time_passed_tot = time_current - self.start_time
                self.time_prev = time_current

                progress_factor = n / self.total
                if progress_factor > 1:
                    progress_factor = 1

                filled_length = int(self.length * progress_factor)
                bar_str = self.fill * filled_length + '-' * (self.length - filled_length)

                fract_str = f'{n}/{self.total}' if self.show_fract else ''

                speed_str = ''
                self.speed_a.upd(n / time_passed_tot)
                self.speed_c.upd((n-self.n_prev) / time_passed_prev)
                speed_a = self.speed_a()
                if self.show_speed_avg:
                    speed_str = self._speed_to_str(speed_a)
                if self.show_speed_cur:
                    speed_str += f'[{self._speed_to_str(self.speed_c())}]'

                eta_str = ''
                if self.show_eta and progress_factor < 1:
                    if speed_a > 0:
                        eta = (self.total - n) / speed_a
                        if eta < 0:
                            eta = 0
                        eta_str = self._time_to_str(eta)
                    else:                 eta_str = '---'
                    eta_str = f'ETA:{eta_str}'

                detailsL = [fract_str,speed_str,eta_str]
                detailsL = [e for e in detailsL if e]
                details_str = f'{" ".join(detailsL)} ' if detailsL else ''

                elapsed_str = ''
                if progress_factor == 1 and self.show_eta:
                    elapsed_str = '-- TOT:' + self._time_to_str(time_passed_tot)
                    elapsed_str += f' {self._speed_to_str(self.total / time_passed_tot)}'

                if self.name:
                    pfx = self.name
                    if prefix:
                        pfx += f' {prefix}'
                else:
                    pfx = prefix

                nfo = f'{pfx}|{bar_str}|{progress_factor * 100:.1f}% {details_str}{elapsed_str}{suffix}'
                printover(nfo)

                if progress_factor == 1:
                    print()
                    if self.logger:
                        self.logger.info(f'ProgBar: {nfo}')

                self.n_prev = n
                self.inc_cached = 0

    def inc(self, prefix:str='', suffix:str='', by:int=1):
        """ increase by """
        self.inc_cached += by
        self(n=self.n_prev + self.inc_cached, prefix=prefix, suffix=suffix)



from abc import ABC, abstractmethod
import numpy as np
from sklearn.svm import SVR
from typing import List, Optional, Tuple, Dict

from pypaq.pytypes import NPL, NUM
from pypaq.lipytools.pylogger import get_pylogger
from pypaq.pms.base import PMSException
from pypaq.pms.paspa import PaSpa
from pypaq.pms.points_cloud import VPoint


class SpaceEstimator(ABC):

    def __init__(
            self,
            logger=         None,
            loglevel=       20):
        if not logger:
            logger = get_pylogger(level=loglevel)
        self.logger = logger

    # extracts X & y from vpoints & space
    @staticmethod
    def _extract_Xy(vpoints:List[VPoint], space:PaSpa) -> Tuple[np.ndarray, np.ndarray]:

        points = [vp.point for vp in vpoints]
        points_normalized = [space.point_normalized(p) for p in points]
        keys = sorted(list(points_normalized[0].keys()))
        points_feats = [[pn[k] for k in keys] for pn in points_normalized]

        scores = [vp.value for vp in vpoints]

        return np.asarray(points_feats), np.asarray(scores)

    # updates model with given data, returns dict with update info
    @abstractmethod
    def update(self, X_new:NPL, y_new:NPL) -> Dict[str,NUM]:
        pass

    # updates model with given VPoints, returns dict with update info
    def update_vpoints(self, vpoints:List[VPoint], space:PaSpa) -> Dict[str,NUM]:
        X, y = SpaceEstimator._extract_Xy(vpoints, space)
        return self.update(X, y)

    # predicts
    @abstractmethod
    def predict(self, X:NPL) -> np.ndarray:
        pass

    def predict_vpoints(self, vpoints:List[VPoint], space:PaSpa) -> np.ndarray:
        X, y = SpaceEstimator._extract_Xy(vpoints, space)
        return self.predict(X)

    # MSE average weighted loss
    @staticmethod
    def loss(
            model,
            y_test: NPL,
            X_test: Optional[NPL]=  None,
            preds: Optional[NPL]=   None,
            weights: Optional[NPL]= None,
    ) -> float:

        if X_test is None and preds is None:
            raise PMSException('\'X_test\' or \'preds\' must be given')

        if preds is None:
            preds = model.predict(X=X_test)

        if type(preds) is not np.ndarray:
            preds = np.asarray(preds)
        if type(y_test) is not np.ndarray:
            y_test = np.asarray(y_test)

        l = (preds - y_test) ** 2

        if weights is not None:
            if type(weights) is not np.ndarray:
                weights = np.asarray(weights)
            l *= weights

        return float(np.mean(l))

    # Estimator status
    @property
    def fitted(self) -> bool:
        return False

    # Estimator state
    @property
    def state(self) -> Dict:
        return {}

    @classmethod
    @abstractmethod
    def from_state(cls, state:Dict, logger):
        pass

    def __str__(self):
        return 'SpaceEstimator'


# Support Vector Regression (SVR) with Radial Basis Function (RBF) kernel based Space Estimator
class RBFRegressor(SpaceEstimator):

    # C & gamma parameters possible values
    VAL = {
        'c': [100, 10, 1, 0.1, 0.01],
        'g': [0.01, 0.1, 1, 10, 100]}

    def __init__(
            self,
            epsilon: float=     0.01,
            num_tests: int=     9,      # number of cross-validation tests while updating
            test_size: float=   0.25,   # size of test data (factor of all)
            y_cut_off: float=   0.1,    # value of y with lower loss (for sparse y)
            seed=               123,
            **kwargs,
    ):

        super().__init__(**kwargs)

        self._indexes =  {'c':0, 'g':0}
        self._epsilon = epsilon

        self._model = None

        self.data_X: Optional[NPL] = None
        self.data_y: Optional[NPL] = None

        self._num_tests = num_tests
        self._test_size = test_size
        self._y_cut_off = y_cut_off

        np.random.seed(seed)

    # builds SVR RBF model from given indexes of parameters
    def _build_model(self, cix:Optional[int]=None, gix:Optional[int]=None) -> SVR:
        if cix is None: cix = self._indexes['c']
        if gix is None: gix = self._indexes['g']
        return SVR(
            kernel=     "rbf",                      # Radial Basis Function kernel
            C=          RBFRegressor.VAL['c'][cix], # regularization, controls error with margin, lower C -> larger margin, more support vectors, longer fitting time
            gamma=      RBFRegressor.VAL['g'][gix], # controls shape of decision boundary, larger value for more complex
            epsilon=    self._epsilon)              # threshold for what is considered an acceptable error rate in the training data

    # fits model, returns test loss
    @staticmethod
    def _fit(
            model,
            X_train: NPL,
            y_train: NPL,
            X_test: NPL,
            y_test: NPL,
            weights: Optional[NPL]= None,
    ) -> float:
        model.fit(X=X_train, y=y_train)
        return RBFRegressor.loss(model=model, X_test=X_test, y_test=y_test, weights=weights)

    # prepares weights for y
    def _weights(self, y:NPL) -> np.ndarray:
        y_min = np.min(y)
        y_max = np.max(y)
        y_cut = y_min + (y_max - y_min) * self._y_cut_off
        factor = 1 - np.mean((y <= y_cut).astype(int))
        factor = max(0.1, factor)
        return np.where(y > y_cut, 1, factor)

    def update(self, X_new:NPL, y_new:NPL) -> Dict[str,NUM]:
        """ concatenates data, tries to update model params, fits & returns some info """

        self.logger.debug(f'preparing for update with {self._num_tests}X cross-validation')

        loss_on_new = 0
        if self.fitted:
            loss_on_new = self.loss(
                model=  self._model,
                X_test= X_new,
                y_test= y_new)

        # add new data
        self.data_X = np.concatenate([self.data_X, X_new]) if self.data_X is not None else X_new
        self.data_y = np.concatenate([self.data_y, y_new]) if self.data_y is not None else y_new

        ### prepare valid future configs

        m_configs = {
            'current':  {'cix': self._indexes['c'],     'gix': self._indexes['g']},     # current
            'cH':       {'cix': self._indexes['c'] + 1, 'gix': self._indexes['g']},     # c higher
            'cL':       {'cix': self._indexes['c'] - 1, 'gix': self._indexes['g']},     # c lower
            'gH':       {'cix': self._indexes['c'],     'gix': self._indexes['g'] + 1}, # g higher
            'gL':       {'cix': self._indexes['c'],     'gix': self._indexes['g'] - 1}} # g lower

        not_valid = []
        for k in m_configs:
            for p in m_configs[k]:
                if m_configs[k][p] in [-1,len(RBFRegressor.VAL[p[0]])]:
                    not_valid.append(k)
        not_valid = list(set(not_valid))
        for k in not_valid:
            m_configs.pop(k)
        self.logger.debug(f'valid alternative models configs for update: {m_configs}')

        models = {config: self._build_model(**m_configs[config]) for config in m_configs}

        acc_loss = {k: 0 for k in m_configs}
        for _ in range(self._num_tests):

            ### prepare data split

            data_size = len(self.data_X)
            train_size = int(data_size * (1-self._test_size))
            choice = np.random.choice(range(data_size), size=train_size, replace=False)
            tr_sel = np.zeros(data_size, dtype=bool)
            tr_sel[choice] = True
            ts_sel = ~tr_sel

            X_train = self.data_X[tr_sel]
            y_train = self.data_y[tr_sel]
            X_test = self.data_X[ts_sel]
            y_test = self.data_y[ts_sel]

            losses = {config: RBFRegressor._fit(
                model=      models[config],
                X_train=    X_train,
                y_train=    y_train,
                X_test=     X_test,
                y_test=     y_test,
                weights=    self._weights(y=y_test)) for config in models}

            for k in losses:
                acc_loss[k] += losses[k]

        for k in acc_loss:
            acc_loss[k] /= self._num_tests
        self.logger.debug(f'loss achieved with each config: {acc_loss}')

        acc_loss_current = acc_loss.pop('current')

        # search for k to update, priority for lower
        update_k = None
        losses_sorted = sorted([(k,l) for k,l in acc_loss.items()], key=lambda x:x[1])
        for e in losses_sorted:
            if e[1] < acc_loss_current:
                update_k = e[0]
                break
        self.logger.debug(f'RBFRegressor is updating: {update_k}')

        if update_k is not None:
            self._indexes[update_k[0]] += 1 if update_k[1] == 'H' else -1
            self._model = models[update_k]
        else:
            self._model = models['current']

        self.logger.debug(f' > current config of indexes: {self._indexes}')

        # finally fit with all data
        loss_all_data_weighted = RBFRegressor._fit(
            model=      self._model,
            X_train=    self.data_X,
            y_train=    self.data_y,
            X_test=     self.data_X,
            y_test=     self.data_y,
            weights=    self._weights(y=self.data_y))

        return {
            'loss_on_new':              loss_on_new,
            'loss_all_data_weighted':   loss_all_data_weighted,
            'c_ix':                     self._indexes['c'],
            'g_ix':                     self._indexes['g']}

    def predict(self, x:NPL) -> np.ndarray:
        if not self.fitted:
            msg = 'RBFRegressor needs to be fitted before predict'
            self.logger.error(msg)
            raise Exception(msg)
        return self._model.predict(X=x)

    @property
    def fitted(self) -> bool:
        return self._model is not None

    # returns object state
    @property
    def state(self) -> Dict:
        return {
            'indexes':      self._indexes,
            'epsilon':      self._epsilon,
            'num_tests':    self._num_tests,
            'test_size':    self._test_size,
            'y_cut_off':    self._y_cut_off}

    # builds object from a state
    @classmethod
    def from_state(cls, state:Dict, logger):
        reg = cls(
            epsilon=    state['epsilon'],
            num_tests=  state['num_tests'],
            test_size=  state['test_size'],
            y_cut_off=  state['y_cut_off'],
            logger=     logger)
        reg._indexes = state['indexes']
        logger.info(f'RBFRegressor build from state: {state}')
        return reg

    def __str__(self):
        c = RBFRegressor.VAL['c'][self._indexes['c']]
        g = RBFRegressor.VAL['g'][self._indexes['g']]
        return f'SVR RBF, C:{c}, gamma:{g}, data:{len(self.data_X)}'


from typing import Sized, Iterable


def chunked(iterable:Sized, size) -> Iterable:
    """ returns a generator of chunks of a (large?) list """
    for i in range(0, len(iterable), size):
        yield iterable[i:i + size]


import numpy as np

from pypaq.exception import PyPaqException


def _double_hinge_np(
        a_value: float,
        b_value: float,
        a_point: float,
        b_point: float,
        point: np.ndarray,
) -> np.ndarray:
    """ double hinge function _/**
    returns:
    - a_value for point <= a_point
    - b_value for point >= b_point
    - linear interpolation from a_value to b_value in range (a_point;b_point) """

    if b_point < a_point or b_point == a_point and a_value != b_value:
        raise PyPaqException('wrong arguments values!')

    r = (point - a_point) / (b_point - a_point)
    v_max, v_min = (b_value, a_value) if b_value > a_value else (a_value, b_value)
    r = np.maximum(v_min, np.minimum(v_max, r))
    return a_value + (b_value - a_value) * r


def double_hinge(
        a_value: float,
        b_value: float,
        a_point: float,
        b_point: float,
        point: int|float|np.ndarray,
) -> float|np.ndarray:
    r = _double_hinge_np(a_value, b_value, a_point, b_point, point)
    if type(point) is not np.ndarray:
        r = float(r)
    return r


import math
import numpy as np
import pandas as pd
from scipy import stats
from typing import Dict, Tuple

from pypaq.pytypes import NPL
from pypaq.lipytools.printout import nice_scin


# min, avg, max
def mam(vals:NPL) -> Tuple[float,float,float]:
    if len(vals): return min(vals), sum(vals) / len(vals), max(vals)
    else:         return 0.0, 0.0, 0.0

# mean, median, std, SEM, h95, min, max, L2norm
def msmx(vals:NPL, use_scin:bool=True) -> Dict:
    arr = np.asarray(vals) if type(vals) is list else vals
    _mean, _median, _std, _min, _max = float(np.mean(arr)), float(np.median(arr)), float(np.std(arr)), float(np.min(arr)), float(np.max(arr))
    _sem = _std / math.sqrt(len(vals))
    _h95 = _sem * stats.t.ppf(0.975, len(vals) - 1) # 0.975 is 1 + 0.95 / 2
    _L2norm = np.linalg.norm(arr, ord=2)
    r = {'min':    _min,
         'mean':   _mean,
         'median': _median,
         'max':    _max,
         'std':    _std,
         'L2norm': _L2norm,
         'sem':    _sem,
         'h95':    _h95}
    r['string'] = ' '.join([f'{k}:{nice_scin(r[k]):8}' for k in r]) if use_scin else (
        f'min:{r["min"]:.3f} '
        f'mean:{r["mean"]:.3f} '
        f'median:{r["median"]:.3f} '
        f'max:{r["max"]:.3f} '
        f'std:{r["std"]:.3f} '
        f'L2norm:{r["L2norm"]:.1f} '
        f'sem:{r["sem"]:.3f} '
        f'h95:{r["h95"]:.3f}')
    return r

# some stats with pandas
def stats_pd(
        val_list :NPL,
        n_percentiles=  10) -> str:
    s = f'{pd.Series(val_list).describe(percentiles=[0.1*n for n in range(1,n_percentiles)])}'
    return s[:s.rfind('\n')]


import matplotlib
from matplotlib import pyplot as plt
import numpy as np
import os
import pandas as pd
import plotly.express as px
import scipy
from typing import List, Optional, Union, Sequence
import warnings

from pypaq.pytypes import NPL
from pypaq.lipytools.files import prep_folder
from pypaq.lipytools.stats import msmx



def histogram(
        val_list: NPL,
        name=                       'values',
        rem_nstd: float=            0.0,    # removes values out of N*stddev
        msmx_stats=                 True,   # prints minimal stats
        density=                    True,
        bins: Optional[int]=        None,   # automatic for None
        add_density_curve: bool=    True,
        save_FD :str=               None,
) -> str:

    if type(val_list) is not np.ndarray:
        val_list = np.asarray(val_list)

    if not len(val_list):
        msg = f'cannot prepare histogram for empty val_list!'
        warnings.warn(msg)
        return msg

    val_set = list(set(val_list))

    small_int_case = False
    if len(val_set) < 50:
        val_set = [v.item() for v in val_set]
        if all([type(v) is int for v in val_set]):
            val_set = sorted(val_set)
            if val_set[-1]-val_set[0] < 60:
                small_int_case = True

    s = []
    if msmx_stats:
        s.append(f'histogram: "{name}" (samples: {len(val_list)}): {msmx(val_list)["string"]}')

    if rem_nstd:
        std = val_list.std()
        mean = val_list.mean()
        val_list = val_list[val_list > mean - rem_nstd * std]
        val_list = val_list[val_list < mean + rem_nstd * std]

    if not bins:
        if small_int_case:
            bins = np.arange(val_set[0]-0.5, val_set[-1]+1.5, 1)
        else:
            bins = len(val_set)
            if bins > 50: bins = 50

    plt.figure()
    n, x, _ = plt.hist(val_list, label=name, density=density, bins=bins, alpha=0.5)
    if small_int_case:
        plt.xticks(range(val_set[0], val_set[-1]+1))

    if add_density_curve:
        # try build density, for some val_list it is not possible
        try:
            density = scipy.stats.gaussian_kde(val_list)
            plt.plot(x, density(x))
        except: pass

    plt.legend(loc='upper right')
    plt.grid(True)
    if save_FD:
        prep_folder(save_FD)
        plt.savefig(f'{save_FD}/{name}.png')
    else:
        plt.show()

    plt.close()
    return '\n'.join(s)


def two_dim(
        y: Union[List,np.ndarray],                  # two(yx) or one(y) dim list or np.array
        x: Optional[Union[List,np.ndarray]]=    None,
        name=                                   'values',
        plot_type: str=                         'plot', # pot, bar, scatter
        save_FD: str =                          None,
        xlogscale=                              False,
        ylogscale=                              False,
        legend_loc=                             'upper left',
        **plot_f_kwargs): # like for scatter: alpha=0.5, s=10

    _plot_type_function = {
        'plot':     plt.plot,
        'bar':      plt.bar,
        'scatter':  plt.scatter}

    if type(y) is list: y = np.asarray(y)
    if x is None:
        if len(y.shape) < 2: x = np.arange(len(y))
        else:
            x = y[:, 1]
            y = y[:, 0]

    plt.figure()
    plot_f = _plot_type_function[plot_type]
    plot_f(x, y, label=name, **plot_f_kwargs)
    if xlogscale: plt.xscale('log')
    if ylogscale: plt.yscale('log')
    plt.legend(loc=legend_loc)
    plt.grid(True)
    if save_FD:
        prep_folder(save_FD)
        plt.savefig(f'{save_FD}/{name}.png')
    else:
        plt.show()
    plt.close()


def two_dim_multi(
        ys: List[Union[List,np.ndarray]],
        names: Optional[List[str]]= None,
        name: Optional[str]=        None,
        save_FD: str=               None,
        xlogscale=                  False,
        ylogscale=                  False,
        legend_loc=                 'upper left'):

    if names is None:
        names = [f'values_{ix}' for ix in range(len(ys))]

    x = np.arange(len(ys[0]))

    plt.figure()
    for y,name in zip(ys,names):
        plt.plot(x, y, label=name)

    plt.legend(loc=legend_loc)
    plt.grid(True)
    if xlogscale: plt.xscale('log')
    if ylogscale: plt.yscale('log')

    if save_FD:
        prep_folder(save_FD)
        if name is None:
            name = "_".join(names)
        plt.savefig(f'{save_FD}/{name}.png')
    else:
        plt.show()
    plt.close()


def three_dim(
    xyz: Union[list, np.ndarray], # sequence of (x,y,val) or (x,y,z,val)
    name=               'values',
    x_name=             'x',
    y_name=             'y',
    z_name=             'z',
    val_name=           'val',
    opacity=            0.7,
    width=              700,
    height=             700,
    save_FD: str =      None):

    # expand to 3 axes + val (3rd axis data)
    if len(xyz[0])<4:
        new_xyz = []
        for e in xyz: new_xyz.append(list(e) + [e[-1]])
        xyz = new_xyz

    df = pd.DataFrame(
        data=       xyz,
        columns=    [x_name,y_name,z_name,val_name])

    std = df[val_name].std()
    mean = df[val_name].mean()
    off = 2*std
    cr_min = mean - off
    cr_max = mean + off

    fig = px.scatter_3d(
        data_frame=     df,
        title=          name,
        x=              x_name,
        y=              y_name,
        z=              z_name,
        color=          val_name,
        range_color=    [cr_min,cr_max],
        opacity=        opacity,
        width=          width,
        height=         height)

    if save_FD:
        file = f'{save_FD}/{name}_3Dplot.html'
        fig.write_html(file, auto_open=False if os.path.isfile(file) else True)
    else:
        fig.show()
    plt.close()


def week_density_plot(
        df: pd.DataFrame,
        bounds: Sequence[float]=    (0,1,3,6,10,20),
        name: Optional[str]=        'density',
        save_FD: Optional[str]=     None,
):
    """plots github-style density calendar
    df: dataframe with 'date' and 'density' columns
    bounds: bin edges"""

    for cn in ['date','density']:
        if cn not in df.columns:
            raise ValueError(f'{cn} not present in dataframe!')

    start = df['date'].min()
    end = df['date'].max()
    start_aligned = start - pd.Timedelta(days=start.weekday())
    end_aligned = end + pd.Timedelta(days=(6 - end.weekday()))

    dates = pd.date_range(start_aligned, end_aligned, freq="D")
    df_extended = pd.DataFrame({'date':dates, 'density':np.zeros(len(dates))})
    df_extended.set_index('date', inplace=True)
    df.set_index('date', inplace=True)
    df_extended.update(df)
    df_extended.reset_index(inplace=True)
    df = df_extended

    df["dow"] = df["date"].dt.dayofweek
    df["week"] = ((df["date"] - start_aligned).dt.days // 7).astype(int)

    matrix = df.pivot(index="dow", columns="week", values="density")

    cmap = matplotlib.colormaps["Oranges"].resampled(len(bounds) - 1)
    norm = matplotlib.colors.BoundaryNorm(bounds, cmap.N)

    cell_in = 0.25  # square size
    cbar_in = 0.25  # colorbar width
    pad_in = 0.10  # gap
    margins_in = (0.5, 0.6, 0.7, 0.3)  # LRBT
    dpi = 100

    h, w = matrix.shape
    left_in, right_in, bottom_in, top_in = margins_in

    img_w_in = w * cell_in
    img_h_in = h * cell_in

    fig_w_in = left_in + img_w_in + pad_in + cbar_in + right_in
    fig_h_in = bottom_in + img_h_in + top_in

    fig = plt.figure(figsize=(fig_w_in, fig_h_in), dpi=dpi)

    def fx(inches):
        return inches / fig_w_in

    def fy(inches):
        return inches / fig_h_in

    ax = fig.add_axes((fx(left_in), fy(bottom_in), fx(img_w_in), fy(img_h_in)))
    im = ax.imshow(matrix, cmap=cmap, norm=norm, origin="lower", interpolation="none", aspect="equal")

    ax.set_yticks(range(h), list(range(h)))

    step = 1
    first_dates = df.groupby("week")["date"].min()
    xtick_positions = np.arange(0, len(first_dates), step)
    xtick_labels = [first_dates.iloc[i].strftime("%b %d") for i in xtick_positions]
    ax.set_xticks(xtick_positions, xtick_labels, rotation=90)

    title = name or ''
    if title:
        title += ', '
    title += f"{start_aligned.strftime('%Y-%m-%d')} - {end_aligned.strftime('%Y-%m-%d')}"
    ax.set_title(title, fontsize=10)

    cax = fig.add_axes((fx(left_in + img_w_in + pad_in), fy(bottom_in), fx(cbar_in), fy(img_h_in)))
    fig.colorbar(im, cax=cax)

    if save_FD:
        prep_folder(save_FD)
        plt.savefig(f'{save_FD}/{name or "density"}.png')
    else:
        plt.show()
    plt.close()